/* burner.c -- Burn the internal EEPROM
   Copyright 2000, 2001, 2002 Free Software Foundation, Inc.
   Written by Stephane Carrez (stcarrez@nerim.fr)

This file is part of GEL.

GEL is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GEL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GEL; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

/*! @page burner Internal Eprom loader and burner.

  This 68HC11 bootstrap program acts as a simple server for uploading
  a program and burn it in the 68HC11 EEPROM.  It supports only two commands:

  M{ADDR-HIGH}{ADDR-LOW}{SIZE}[{DATA}]

     Write the memory at given address.


  G{ADDR-HIGH}{ADDR-LOW}

     Jump to program at given address.


  This bootstrap program is uploaded by GTAM to be able to burn
  a program in the internal EEPROM.

  Implementation Notes:

     -# This program must not be linked with a startup file.  It implements
        the startup entry point.
     -# The \c _start function must be at beginning of this file.
        By doing so, it will be mapped at address 0 by the linker and
        we avoid to have some jump to call it (since the boot will
        jump there).
     -# It must be compiled with \b -mshort \b -Os \b -fomit-frame-pointer
        to make sure it fits in less than 240 bytes (keep a few bytes at
        end for the stack).
     -# It must be compiled with \b -msoft-reg-count=0 or =1 to reduce the
        use of soft-registers (since they are mapped in RAM in .page0).
     -# The pointer used to write the memory is marked \b volatile to prevent
        GCC from optimizing memory accesses.  We rely on this for the good
        computation of the CRC.
     -# Before scanning the memory, we switch the bus in the expanded mode.
        We do not switch back to single chip when jumping to the program.
     -# The soft registers must be located at beginning of the .page0.
        This will clobber the beginning of the \c _start function.
        We don't care because this clobbers some initialisation part.

  Caveats:
  <ul>
     <li>It is not possible to write in the 68HC11 RAM.  We don't check
         for this and this will probably crash.
  </ul>

  @htmlonly
  Source file: <a href="burner_8c-source.html">burner.c</a>
  @endhtmlonly

*/

#include <sys/ports.h>

typedef void __attribute__ ((noreturn)) (* func)();

static unsigned short get_char (void);
static unsigned char* get_addr (void);
volatile int __attribute__((noreturn)) main ();
static void flush (void);
void eeprom_write_byte (unsigned char *p, const unsigned char value);
void udelay_10ms (void);
void _start (void);

void
_start()
{
  __asm__ __volatile__ ("bra main");
  /* main (); */
}

static void
flush ()
{
  while (!(_io_ports[M6811_SCSR] & M6811_TDRE))
    continue;
}

static void
put_char (unsigned char c)
{
  flush ();
  _io_ports[M6811_SCDR] = c;
  _io_ports[M6811_SCCR2] |= M6811_TE;
}

volatile int
main ()
{
  unsigned char* addr;
  unsigned char c;
  
  while (1)
    {
      /* Print banner for synchronization.  */
      put_char ('\n');
      put_char ('>');

      /* Wait for command.  */
      c = get_char ();

      /* Write memory command.  Command format is:

         M<ADDR-HIGH><ADDR-LOW><SIZE>[<DATA>]

         Address, size and data are passed in binary form.
         A size of 0 corresponds to 256 bytes.

	 Since burning a byte can take some time, we must synchronize
	 with the host.  To do this, we echo the byte that was programmed
	 and the host waits for it before sending the next one.  If the
	 host does not wait, we can loose some bytes.  */
      if (c == 'M')
        {
          unsigned char size;
          
          addr = get_addr ();
          size = get_char ();
          do
            {
              *addr = get_char ();
	      eeprom_write_byte (addr, get_char ());
	      put_char (*addr);
            }
          while (--size != 0);
        }

      /* Go command.  Command format is:

         G<ADDR-HIGH><ADDR-LOW>

         We reply with a G.  Depending on the program, the final \n
         may not be received since we don't wait for it to be sent.  */
      else if (c == 'G')
        {
          func handler;
            
          addr = get_addr ();
          put_char ('G');
          put_char ('\n');
          flush ();

          handler = (func) addr;
          handler ();
        }

      /* For others, emit something to tell we are alive.  */
      else
        {
          put_char ('?');
        }
    }
}

void
udelay_10ms ()
{
  unsigned short tmp = 40000;

  __asm__ __volatile__ ("1: subd #1\n"
			"	bne 1b" : "=d"(tmp) : "d"(tmp));
}

/*! Write a byte in eeprom.

    Write the byte \b value in the eeprom at address \b p.
    The eeprom byte location is erased first if write operation
    is not possible (when at least one bit must be set to 1).

    @param p      Address of the eeprom byte to write
    @param value  Value to write
    @see eeprom_write_short  */
void 
eeprom_write_byte (unsigned char *p, const unsigned char value)
{
  unsigned char diff;

  diff = *p ^ value;
  if (diff)
    {
      if (diff & value)
	{
	  /* Clear eeprom byte.  */
	  diff = *p ^ value;
	  if (diff == 0)
	    return;
	  
	  _io_ports[M6811_PPROG] = M6811_EELAT | M6811_ERASE | M6811_BYTE;
	  *p = value;
	  _io_ports[M6811_PPROG] = M6811_EELAT | M6811_ERASE
	    | M6811_BYTE | M6811_EEPGM;

	  /* Wait 10 ms.  */
	  udelay_10ms ();
	  _io_ports[M6811_PPROG] = M6811_EELAT | M6811_ERASE | M6811_BYTE;
	}
      
      /* Write eeprom byte.  */
      _io_ports[M6811_PPROG] = M6811_EELAT;
      *p = value;
      _io_ports[M6811_PPROG] = M6811_EELAT | M6811_EEPGM;

      /* Wait 10 ms.  */
      udelay_10ms ();
      _io_ports[M6811_PPROG] = M6811_EELAT;
      _io_ports[M6811_PPROG] = 0;
    }
}

static unsigned char*
get_addr ()
{
  unsigned short c1, c2;

  c1 = get_char ();
  c2 = get_char ();
  return (unsigned char*) (((unsigned short) c1 << 8) | ((unsigned short) c2));
}

static unsigned short
get_char ()
{
  unsigned char c;
  volatile unsigned char* ports = &_io_ports[0];
  
  while (1)
    {
      c = ports[M6811_SCSR];

      if (c & M6811_RDRF)
        break;
    }
  return ports[M6811_SCDR];
}
