/* timer.c -- Timer
   Copyright 2000, 2001, 2002 Free Software Foundation, Inc.
   Written by Stephane Carrez (stcarrez@worldnet.fr)

This file is part of GEL.

GEL is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GEL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GEL; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include <sys/param.h>
#include <sys/locks.h>
#include <sys/interrupts.h>
#include <sys/ports.h>
#include <gel/timer.h>

/** Timer list structure.

    Timers are kept in a ring.

*/
struct timer_list 
{
  struct timer *first;
  struct timer *last;
  unsigned short flags;
  long timeout;
  long ticks_to_wait;
  unsigned char dispatching;
};

static struct timer_list fast_timers PAGE0_ATTRIBUTE;
static struct timer_list slow_timers PAGE0_ATTRIBUTE;

static unsigned long time_usec PAGE0_ATTRIBUTE;
static unsigned long time_sec PAGE0_ATTRIBUTE;

static unsigned long boot_sec;
static long total_adjtime;
static long time_adj_usec;
unsigned long _timer_current_overflow PAGE0_ATTRIBUTE;

timer_overflow_handler _overflow_handler PAGE0_ATTRIBUTE = 0;

#define TICK_USEC (unsigned long) (TIMER_TICK_VALUE (12 + (1 << M6811_DEF_RTR)))

#define TIMER_OVERFLOW_TO_USEC \
  ((65536 * 4 * (1 << M6811_DEF_TPR)) / (M6811_CPU_CLOCK / 1000000L))

/* Execute the timer handlers which have ellapsed.


   Return 1 if there is no more pending timers and return 0 if the
   timer list is not empty.  */
static int
timer_dispatch (struct timer_list *tlist)
{
  struct timer *t;

  /* Lock the dispatcher to prevent recursion.  */
  tlist->dispatching++;

  while (tlist->ticks_to_wait <= 0)
    {
      t = tlist->first;
      if (t->flags & TIMER_HEAD)
        {
          tlist->dispatching--;
          return 1;
        }

      tlist->ticks_to_wait += t->next->timeout;

      /* Remove the timer from the list before processing it.  */
      t->next->prev = t->prev;
      t->prev->next = t->next;
      t->next = 0;
      t->prev = 0;

      /* Unmask the interrupts for slow callbacks.  */
      if (t->flags & TIMER_UNMASK)
	{
	  unlock ();
	  t->handler (t);
	  lock ();
	}
      else
	{
          interruption_point ();
	  t->handler (t);
	}
    }
  tlist->dispatching--;
  return 0;
}

/** Timer interrupt handler.

 */
void __attribute__((interrupt))
timer_interrupt (void)
{
  struct timer_list *tlist = &fast_timers;
  
  /* Acknowledge timer interrupt and do timer dispatching.
     Nested interrupts can happen within timer_dispatch().  When this
     happens, we keep track of the nested timer interrupt in the
     'ticks_to_wait' member and we don't re-enter in the dispatcher.  */
  _io_ports[M6811_TFLG2] = M6811_RTIF;

  tlist->ticks_to_wait--;
  if (tlist->ticks_to_wait <= 0 && tlist->dispatching == 0)
    {
      if (timer_dispatch (tlist) != 0)
	{
	  /* When the timer list becomes empty,
	     disable the timer interrupt.  */
	  _io_ports[M6811_TMSK2] &= ~M6811_RTII;
	}
    }
}

/** Timer overflow interrupt handler.

    It is defined to increment the overflow counter.

*/
void __attribute__((interrupt))
timer_overflow_interrupt (void)
{
  /* Acknowledge the timer overflow interrupt.  */
  _io_ports[M6811_TFLG2] = M6811_TOF;

  /* Keep track of the number of overflow that occurred.  */
  _timer_current_overflow++;

  /* Call the timer overflow application handler.  */
  if (_overflow_handler)
    _overflow_handler ();

  /* Update the time variables.  */
  time_usec += TIMER_OVERFLOW_TO_USEC;
  if (time_usec >= 1000000L)
    {
      struct timer_list *tlist = &slow_timers;
      
      time_usec -= 1000000L;
      time_sec++;

      /* Dispatch slow timers.  */
      tlist->ticks_to_wait--;
      if (tlist->ticks_to_wait <= 0 && tlist->dispatching == 0)
        {
          timer_dispatch (tlist);
        }
    }
}

/*! Get the current time.

    Returns in the \b tv parameter the current time.  The current
    time is expressed in seconds and microseconds.

    The current time is computed by taking a snapshot of a
    timer overflow counter and the free running counter.  A boot
    time is added to the second portion to obtain the current time.
    Accurracy of the result depends on the resolution of the free
    running counter.  The value returned is also a snapshot of the
    time within \b timer_gettime.

    This function is intended to be close to the Unix \b gettimeofday
    operation.

    @param tv pointer to the timeval structure for the result
    @see timer_settime, timer_adjtime  */
void
timer_gettime (struct timeval *tv)
{
  unsigned short mask;
  unsigned short tcnt;
  unsigned long usec;
  
  /* Get the time computed with the counter overflow and
     get the free running counter.  */
  mask = lock ();
  tv->tv_sec = time_sec;
  usec = time_usec;
  tcnt = get_timer_counter ();

  /* If there is a pending overflow, take it into account.  */
  if (_io_ports[M6811_TFLG2] & M6811_TOF)
    {
      /* If the overflow occurred after we read the timer counter,
         ignore it.  Otherwise, take it into account and update usec.  */
      if (tcnt < get_timer_counter ())
        usec += TIMER_OVERFLOW_TO_USEC;
    }
  restore (mask);

  /* Translate into seconds and microseconds.  */
  usec += tcnt_to_usec (tcnt);
  if (usec >= 1000000L)
    {
      usec -= 1000000L;
      tv->tv_sec ++;
    }
  tv->tv_usec = usec;
}

/*! Set the current time.

    Set the current time to \b tv.

    This function is intended to be close to the Unix \b settimeofday
    operation.

    @param tv  pointer to the timeval structure holding the new time
    @see timer_gettime, timer_adjtime  */
void
timer_settime (struct timeval *tv)
{
  unsigned short mask;
  
  mask = lock();
  time_sec = tv->tv_sec;
  restore (mask);
}

/* Timer callback to adjust the time periodically and smoothly.  */
static void
timer_adjtime_callback (struct timer *t)
{
  if (total_adjtime < 0)
    {
      time_usec -= time_adj_usec;
      total_adjtime += time_adj_usec;
      if (total_adjtime > 0)
	{
	  time_usec += total_adjtime;
	  total_adjtime = 0;
	}
    }
  else
    {
      time_usec += time_adj_usec;
      total_adjtime -= time_adj_usec;
      if (total_adjtime < 0)
	{
	  time_usec -= total_adjtime;
	  total_adjtime = 0;
	}
    } 
  if (time_usec >= 1000000L)
    {
      time_usec -= 1000000L;
      time_sec ++;
    }
  else if (time_usec < 0)
    {
      time_usec += 1000000L;
      time_sec --;
    }
  if (total_adjtime)
    {
      t->timeout = 1;
      timer_insert (t);
    }
}

long
timer_adjtime (long adj_usec)
{
  static struct timer adj_timer
    = TIMER_INIT_STATIC (timer_adjtime_callback, TIMER_MASK);

  unsigned short mask;
  long old_adj;
  
  mask = lock ();
  old_adj = total_adjtime;
  total_adjtime = adj_usec;
  if (timer_is_active (&adj_timer) == 0)
    {
      restore (mask);
      adj_timer.timeout = 1;
      timer_insert (&adj_timer);
    }
  else
    {
      restore (mask);
    }
  return old_adj;
}

#define TIMER_TICK_VALUE(DIV) \
  (((1L << (DIV)) * 1000L) / (M6811_CPU_E_CLOCK / 1000L))

static void
timer_list_create (struct timer_list *tlist)
{
  tlist->first = (struct timer*) tlist;
  tlist->last = (struct timer*) tlist;
  tlist->dispatching = 0;
  tlist->ticks_to_wait = 0;
  tlist->timeout = 0x7fffffffL;
  tlist->flags = TIMER_HEAD;
}

void
timer_initialize (void)
{
  unsigned short mask;
  
  /* Create two circular list of timers: one for fast timers (RTI based)
     and one for slow timers (1 sec based).  */
  timer_list_create (&fast_timers);
  timer_list_create (&slow_timers);

  time_sec = 0;
  time_usec = 0;
  _timer_current_overflow = 0;
  boot_sec = 0;

  /* Timer overflow interrupt is always enabled.
     Timer interrupt is enabled only when a fast timer exist.  */
  mask = lock ();
  _io_ports[M6811_TMSK2] = M6811_TOI;

  set_interrupt_handler (RTI_VECTOR, timer_interrupt);
  set_interrupt_handler (TIMER_OVERFLOW_VECTOR, timer_overflow_interrupt);
  
  restore (mask);
}

void
timer_create (struct timer *t, unsigned long timeout, timer_handler handler,
	      unsigned char flags)
{
  t->handler = handler;
  t->timeout = timeout;
  t->flags = flags;
  timer_insert (t);
}

void
timer_insert (struct timer *t)
{
  unsigned short mask;
  struct timer* pt;
  long timeout;
  struct timer_list *tlist;
  
  if (timer_is_active (t))
    timer_remove (t);

  /* Insert the new timer in the slow timer list.  In that case,
     the timeout is expressed in seconds and no computation is necessary.  */
  if (t->flags & TIMER_SLOW)
    {
      tlist = &slow_timers;
      timeout = t->timeout;
    }
  else
    {
      tlist = &fast_timers;
      timeout = (t->timeout / TICK_USEC) * 1000L;
      timeout += (1000L * (t->timeout % TICK_USEC) + TICK_USEC - 1)
        / TICK_USEC;
    }
  
  mask = lock ();
  /* Must adjust the timeout to take into account the pending
     ticks we will process soon.  */
  if (tlist->ticks_to_wait < 0)
    {
      timeout += -tlist->ticks_to_wait;
    }
  else if (tlist->ticks_to_wait > timeout)
    {
      if (tlist->first != (struct timer*) tlist)
	tlist->first->timeout = tlist->ticks_to_wait;
    }
  else if (tlist->first != (struct timer*) tlist)
    timeout += tlist->first->timeout - tlist->ticks_to_wait;
  
  for (pt = tlist->first; pt != (struct timer*) tlist; pt = pt->next)
    {
      timeout -= pt->timeout;
      if (timeout < 0)
        {
	  /* New timer is before 'pt'.  */
          timeout += pt->timeout;
          t->timeout = timeout;

          if (pt == tlist->first && tlist->ticks_to_wait > 0)
            tlist->ticks_to_wait = timeout;
          pt->timeout -= timeout;

	  /* Insert before 'pt'.  */
          t->next = pt;
          t->prev = pt->prev;
          pt->prev = t;
          t->prev->next = t;
          restore (mask);
          return;
        }
    }

  /* Insert at end of the list.  */
  t->timeout = timeout;
  t->prev = tlist->last;
  t->prev->next = t;
  t->next = (struct timer*) tlist;
  tlist->last = t;

  /* If this is the first timer inserted in the list,
     activate the dispatching.  */
  if (tlist->first == t)
    {
      tlist->ticks_to_wait = timeout;
      if (tlist == &fast_timers)
        {
          _io_ports[M6811_TMSK2] |= M6811_RTII;

          /* Be sure the RTIF flag is cleared.  */
          _io_ports[M6811_TFLG2] = M6811_RTIF;
        }
    }
  restore (mask);
}

/*! Remove the timer.

  @param t pointer to timer object
  @see timer_insert  */
void
timer_remove (struct timer *t)
{
  unsigned short mask;
  struct timer_list *tlist;

  /* Timer object is already removed.  */
  if (t->next == 0)
    return;

  /* Get the list */
  if (t->flags & TIMER_SLOW)
    tlist = &slow_timers;
  else
    tlist = &fast_timers;
  
  mask = lock ();
  /* When removing the first timer, adjust the ticks to wait.  */
  if (tlist->first == t)
    {
      /* Take into account the ticks that were processed.  */
      t->timeout = tlist->ticks_to_wait;
    }

  if (t->next != (struct timer*) tlist)
    {
      /* Accumulate time to the next timer.  */
      t->next->timeout += t->timeout;
      if (tlist->first == t && tlist->ticks_to_wait > 0)
	tlist->ticks_to_wait = t->next->timeout;
    }
  else if (t->prev == (struct timer*) &fast_timers)
    {
      /* Disable timer interrupts if there is no timer to activate.  */
      _io_ports[M6811_TMSK2] &= ~M6811_RTII;
    }
  t->next->prev = t->prev;
  t->prev->next = t->next;
  restore (mask);
  t->next = 0;
  t->prev = 0;
}

void
timevalfix (struct timeval *tv)
{
  if (tv->tv_usec < 0)
    {
      tv->tv_sec--;
      tv->tv_usec += 1000000;
    }
  if (tv->tv_usec >= 1000000)
    {
      tv->tv_sec++;
      tv->tv_usec -= 1000000;
    }
}

/*! Add a timeval.

    Add in \b to the time in \b val.

    @param to  result and value to add
    @param val value to add
    @see timevalsub  */
void
timevaladd (struct timeval *to, struct timeval *val)
{
  to->tv_sec += val->tv_sec;
  to->tv_usec += val->tv_usec;
  timevalfix (to);
}

/*! Subtract a timeval.

    Subtract from \b to the time in \b val.  The result time can
    become negative.

    @param to  result and value from which to subtract
    @param val value to subtract
    @see timevaladd  */
void
timevalsub (struct timeval *to, struct timeval *val)
{
  to->tv_sec -= val->tv_sec;
  to->tv_usec -= val->tv_usec;
  timevalfix (to);
}

