/* var.c -- Control Variables
   Copyright 2001, 2002 Free Software Foundation, Inc.
   Written by Stephane Carrez (stcarrez@worldnet.fr)

This file is part of GEL.

GEL is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GEL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GEL; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include <sys/param.h>
#include <sys/locks.h>
#include <stddef.h>
#include <gel/var.h>
#include <eeprom.h>
#include <assert.h>
#include <stdio.h>

#ifndef GEL_MAX_VARS
#  define GEL_MAX_VARS 128
#endif

unsigned var_count;
const struct var_def *var_list[GEL_MAX_VARS];

/*! Return the integer value of the variable.

    @param var  Pointer to the variable definition
    @return     Value of the variable (expressed as a long)  */
unsigned long
var_value (const struct var_def *var)
{
  switch (var->type)
    {
    case V_UINT8:
      return (unsigned long) (var->u.uint8[0]);

    case V_UINT16:
      return (unsigned long) (var->u.uint16[0]);

    case V_UINT32:
      return (unsigned long) (var->u.uint32[0]);

    default:
      return 0;
    }
}
      
/*! Format the variable value in the buffer.

    Create a printable representation of the variable.
    The variable is formated according to its format type.

    @param buf  Buffer large enough to hold the result
    @param var  Pointer to the variable definition
    @see var_value  */
void
var_format (char *buf, const struct var_def *var, unsigned long value)
{
  switch (var->format)
    {
    case F_INTEGER:
      sprintf (buf, "%ld", value);
      break;

    case F_ONOFF:
      sprintf (buf, "%s", value ? "ON" : "OFF");
      break;

    default:
      assert (1 == 0);
      break;
    }
}

/*! Set the variable.

    Set the integer variable to a given value.
    The value is truncated to fit in the [min..max] range defined
    by the variable definition.  If the variable lies in EEPROM,
    the EEPROM is written, thus saving the value permanently.

    @param var   Pointer to the variable definition
    @param value Value to assign to the variable
    @see var_format, var_value  */
void
var_set_integer (const struct var_def *var, unsigned long value)
{
  union var_value v;
  unsigned char *addr;
  unsigned char *svalue;
  size_t size;

  if (value <= var->min_value)
    value = var->min_value;
  if (value >= var->max_value)
    value = var->max_value;

  switch (var->type)
    {
    case V_UINT8:
      v.u8 = value;
      svalue = (unsigned char*) &v.u8;
      addr = (unsigned char*) var->u.uint8;
      size = 1;
      break;

    case V_UINT16:
      v.u16 = value;
      svalue = (unsigned char*) &v.u16;
      addr = (unsigned char*) var->u.uint16;
      size = 2;
      break;

    case V_UINT32:
      v.u32 = value;
      svalue = (unsigned char*) &v.u32;
      addr = (unsigned char*) var->u.uint32;
      size = 4;
      break;

    default:
      assert (1 == 0);
      return;
    }
  while (size > 0)
    {
      if (addr >= &eeprom_low && addr < &eeprom_high)
        {
          eeprom_write_byte (addr, *svalue);
        }
      else
        {
          *addr = *svalue;
        }
      addr++;
      svalue++;
      size--;
    }   
}

/*! Register a variable definition.

    @param var  Pointer to the variable definition
    @see var_value, var_format  */
void
var_register (const struct var_def *var)
{
  unsigned long v;
  
  assert (var_count < GEL_MAX_VARS);
  assert (var->def_value >= var->min_value);
  assert (var->def_value <= var->max_value);

  var_list[var_count++] = var;

  /* Verify the current value.  If it is out of bounds, reset
     to the default value.  If the variable is in EEPROM and the
     EEPROM is not initialized, the default value can be initialized
     automatically (provided the max is not 0xff, 0xffff or 0xffffffff).  */
  v = var_value (var);
  if (v < var->min_value || v > var->max_value)
    {
      var_set_integer (var, var->def_value);
    }
}

